/* ---------------------------------------------------------------------- 
 *
 * sm.c
 *
 * calculate R.A., DEC, azimuth and altitude for the Sun and Moon
 *
 * 2004-07-20   Christoph C. Birk (birk@ociw.edu)
 *
 * credit to Ken Clardy for all the geometry
 *
 * compile as object (sm.o):       cc -c sm.c
 * compile as executable (a.out):  cc -DMAKE_EXEC sm.c -lm
 *
 * ---------------------------------------------------------------------- */
 
#define _REENTRANT                /* gmtime_r() */

/* DEFINEs -------------------------------------------------------------- */

#define DEBUG          0

#define PI2            (2.0*M_PI)
#define RADS           (180.0/M_PI)
#define JD_BASE        2415384.5
#define JULIAN_EPOCH   1721045.0
#define JULIAN_YEAR    365.25
#define STANDARD_EPOCH 2000.0
#define JDEPOCH(epoch) (JULIAN_EPOCH+(epoch)*JULIAN_YEAR)

#ifdef MAKE_EXEC
#define DEF_LONG       (-118.16)
#define DEF_LAT        (+34.13)
#endif

/* INCLUDEs ------------------------------------------------------------- */

#ifdef MAKE_EXEC
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#else
#if (DEBUG > 1)
#include <stdio.h>
#endif
#endif
#include <time.h>
#include <math.h>

/* TYPEDEFs ------------------------------------------------------------- */

typedef struct  abstime_tag {
  double  day;
  int     year;
} AbsTime;

/* STATICs -------------------------------------------------------------- */

static int  mlen[12] = { 31,28,31,30,31,30,31,31,30,31,30,31 };
static int  doref=1;      /* apparent position (refraction) */

/* function prototype(s) */
 
static void   normabs     (AbsTime*);
static void   toabs       (AbsTime*,int,int,int,int,int,int);
static double juldat      (AbsTime*);
       double get_jd      (time_t);
static void   vmake       (double*,double,double,double);
static double dot         (double*,double*,int);
static double vdot        (double*,double*);
static void   dif         (double*,double*,int);
static void   vdif        (double*,double*);
static double enorm       (double*,int);
static double cnorm       (double*);
static double vnorm       (double*);
static void   nxs         (double*,double,int);
static void   vxs         (double*,double);
static void   sphere      (double*,double*,double*);
static void   vrot        (double*,double,int);
// static void   copy        (double*,double*,int);
// static void   vcopy       (double*,double*);
static double refract     (double);
static double square      (double);

static void   get_times   (time_t,double,double*,double*,double*,double*);
static void   get_solar_v (double,double,double*);
static void   get_lunar_v (double,double,double,double,double*);

/* GLOBALs -------------------------------------------------------------- */

void get_sun  (time_t,double,double,double*,double*,double*,double*);
void get_moon (time_t,double,double,double*,double*,double*,double*,double*);

/* ---------------------------------------------------------------------- */
/* M A I N */
/* ---------------------------------------------------------------------- */

#ifdef MAKE_EXEC
int main(int argc,char** argv)
{
  int    i,wait=0;
  time_t time_value;
  double longi=DEF_LONG,lati=DEF_LAT;
  double s_ra,s_dec,s_az,s_alt;
  double m_ra,m_dec,m_az,m_alt,m_phase;

  {                                             /* parse command line */
    extern char *optarg;
    extern int opterr,optopt,optind; opterr=0;
    while ((i=getopt(argc,argv,"wc:r")) != EOF) {
      switch (i) {
      case 'w':                                 /* wait for rise/set */
        wait = 1;
        break;
      case 'c':                                 /* coordinates */
        { double a,b;
          if (sscanf(optarg,"%lf %lf",&a,&b) == 2) { 
            longi = a; lati = b;
          }
        }
        break;
      case 'r':                                 /* refraction */
        doref = 1;
        break;
      case '?':
        fprintf(stderr,"%s: option '-%c' unknown or parameter missing\n",
                argv[0],(char)optopt);
        exit(1);
      }
    }
  }

  time(&time_value);                      /* current time */
  get_sun(time_value,longi,lati,&s_ra,&s_dec,&s_alt,&s_az);
  if (wait) { double alt=s_alt;
    fprintf(stdout,"Sun%s at ",(alt < 0.0) ? "Rise" : "Set");
    do {                                  /* wait for sign change */
      time_value += 1+(time_t)fabs(s_alt);
      get_sun(time_value,longi,lati,&s_ra,&s_dec,&s_alt,&s_az);
    } while (alt*s_alt > 0.0);   /* WARNING: sun-diameter = 34 min */
  }
  fprintf(stdout,"lat=%f, long=%f\n",lati,longi);
  fprintf(stdout,"%s",ctime(&time_value));
  fprintf(stdout,"R.A.(sun)=%f, DEC(sun)=%f\n",s_ra,s_dec);
  fprintf(stdout,"AZ(sun)=%f, ALT(sun)=%f\n\n",s_az,s_alt);

  time(&time_value);                      /* current time */
  get_moon(time_value,longi,lati,&m_ra,&m_dec,&m_alt,&m_az,&m_phase);
  if (wait) { double alt=m_alt;
    fprintf(stdout,"Moon%s at ",(alt < 0.0) ? "Rise" : "Set");
    do {                                  /* wait for sign change */
      time_value += 1+(time_t)fabs(m_alt);
      get_moon(time_value,longi,lati,&m_ra,&m_dec,&m_alt,&m_az,&m_phase);
    } while (alt*m_alt > 0.0);         /* WARNING:  moon-diameter != 0 */
  }
  fprintf(stdout,"%s",ctime(&time_value));
  fprintf(stdout,"R.A.(moon)=%f, DEC(moon)=%f\n",m_ra,m_dec);
  fprintf(stdout,"AZ(moon)=%f, ALT(moon)=%f\n",m_az,m_alt);
  fprintf(stdout,"PH(moon)=%f (%.3f,%.3f)\n\n",m_phase,m_phase/180.0,
          pow(m_phase/180.0,2.50));

  return(0);
}
#endif /* MAKE_EXEC */

/* ---------------------------------------------------------- */
/* local functions */
/* ---------------------------------------------------------- */

static void normabs (AbsTime* ab)
{                     /* Normalize an absolute time structure */
  double j;

  j = 1.0 + ((ab->year%4 == 0) ? 366.0 : 365.0);
  while (ab->day >= j)  {
    ab->day -= j;
    (ab->year)++;
    j = 1.0 + ((ab->year%4 == 0) ? 366.0 : 365.0);
  }
  while (ab->day < 1.0)  {
    (ab->year)--;
    j = (ab->year%4 == 0) ? 366.0 : 365.0;
    ab->day += j;
  }
}

/* ---------------------------------------------------------------------- */

static void toabs(AbsTime* ab,int yr,int mo,int dy,int hr,int mn,int sc)
{                                /* Put time into absolute time structure */
  int  m;

  ab->year = yr;

  mlen[1] = (yr%4 == 0) ? 29 : 28;
  for (ab->day=0.0,m=0; m<mo; m++)  ab->day += (double)mlen[m];
  ab->day += (double)dy+((double)hr+((double)mn+(double)sc/60.0)/60.0)/24.0;
  normabs(ab);
}

/* ---------------------------------------------------------------------- */

/* -- Limitation -- implements only an approximate Julian calendar,
   correct for 1 January 1901 to 28 Feb. 2100.  This approximation
   matches the normalization done in the absolute time subroutines.  */

static double juldat (AbsTime* ab)
{                                   /*  (Full) Julian date from absolute  */
  double f;

  f = (double) (      /* Get year and leap year days */
      (1461 * (long int)(ab->year - 1901) )/4
      + ab->day       /* Add the current day of year */
      )
      + JD_BASE;      /* Base time = 0h GMT, 31 Dec, 1900 */
  return(f);
}

/* ---------------------------------------------------------------------- */

double get_jd(time_t ut)
{
  int       thisyear;
  double    jd;
  AbsTime   utc;
  struct tm time_structure;

  gmtime_r(&ut,&time_structure);

  thisyear = time_structure.tm_year;
  if (thisyear < 70) thisyear += 100;
  if (thisyear < 1900) thisyear += 1900;
  toabs (&utc, thisyear,
         time_structure.tm_mon,     time_structure.tm_mday,
         time_structure.tm_hour,    time_structure.tm_min,
         time_structure.tm_sec);

  jd = juldat(&utc);  

  return(jd);
}

/* ---------------------------------------------------------------------- */

static void vmake(double* v,double x,double y,double z) 
{                                      /* Construct a vector */
  v[0] = x;
  v[1] = y;
  v[2] = z;
}

/* ---------------------------------------------------------------------- */

static double dot(double* a,double* b,int n)
{                                      /* N-Vector inner product */
  int    i;
  double sum;

  for (i=0,sum=0.0; i<n; i++)  sum += a[i]*b[i];

  return sum;
}

static double vdot (double* a,double* b)
{                                      /* 3-Vector dot product */
  return dot(a, b, 3);
}

/* ---------------------------------------------------------------------- */

static void dif (double* a,double* b,int n)
{                                      /* N-Vector difference */
  int  i;

  for (i=0; i<n; i++)  a[i] -= b[i];
}

static void vdif (double* a,double* b)
{                                     /* 3-Vector difference */
  dif (a, b, 3);
}

/* ---------------------------------------------------------------------- */

static double enorm (double* a,int n)
{                                     /* Euclidean norm of vector */
  return sqrt(dot(a,a,n));
}

static double cnorm (double* a)
{                                     /* 2-Vector norm */
  return enorm(a,2);
}

static double vnorm (double* a)
{                                     /* 3-Vector norm */
  return enorm (a,3);
}

/* ---------------------------------------------------------------------- */

static void nxs (double* v,double x,int n)
{                                     /* N-Vector by scaler multiply */
  int  i;

  for (i=0; i<n; i++)  v[i] *= x;
}

static void vxs (double* v,double x)
{                                     /* 3-Vector by scaler multiply */
  nxs (v, x, 3);
}

/* ---------------------------------------------------------------------- */

static void sphere(double* a,double* d,double* v) 
{                             /* Obtain spherical coordinates from vector */
  *a = atan2(v[1], v[0]);
  if (*a < 0.0) *a += PI2;
  *d = atan2(v[2], cnorm(v));
}

/* ---------------------------------------------------------------------- */

/* Rotate the 3-vector "v" about the "n"-th axis of a right-handed
   coordinate system by the angle "theta" radians.
   Note that this is equivalent to a rotation of the
   coordinate system by the negative of "theta".     */

static void vrot (double* v,double theta,int n)
{                                     /* Elementary rotation */
  double st,ct,t;
  int    j,k;

  if (theta != 0.0)  {
    j = n % 3;
    k = (j+1) % 3;
    ct = cos (theta);
    st = sin (theta);
    t = ct*v[j] - st*v[k];
    v[k] = ct*v[k] + st*v[j];
    v[j] = t;
  }
}

/* ---------------------------------------------------------------------- */

#if 0 // unused
static void copy(double* v,double* s,int n)
{                                     /* N-Vector copy */
  int     i;

  for (i=0; i<n; i++) v[i] = s[i];
}
#endif

#if 0 // unused
static void vcopy(double* v, double* s)
{                                     /* 3-Vector copy */
  copy(v, s, 3);
}
#endif

/* ---------------------------------------------------------------------- */

/* Correct altitude a (degrees) for refraction */
/* Formula from (new) Explanatory Supplement to A.E. */

#define  RF  0.0167           /* Refraction coefficient in degrees */

static double refract(double a)
{
  double  b, x;
  /* Formula is 0.0167 / tan (a + 8.6/(a + 4.4))  */
  b = a + 4.4;
  x = b > 2.0 ? a + 8.6/b : -0.5 - a;
  /* The limit of b=2, a=-2.4 is arbitrary, and is selected for ease
     of computation.  It is less than the minimum x at a=-1.46742 and
     avoids the singularity at a=-4.4 */
  /*  b = x < 45.0 ? RF / tan(x/RADS) : RF * tan( (90.0-x)/RADS); */
  /*  b = RF * tan ( (90.0 - x) / RADS );  */
  b = doref ? RF * tan ( (90.0 - x) / RADS ) : 0.0;
  return(a+b);
}

#undef RF

/* ---------------------------------------------------------------------- */

static double square(double a)
{
  return(a*a);
}

/* ---------------------------------------------------------- */

static void get_times(time_t t,double lon,
                      double* toy,double* gha,double* ob,double* lst)
{
  int       thisyear;
  double    fjd,gmt,dum,sml,gst;
  AbsTime   utc;
  struct tm time_structure;

  gmtime_r(&t,&time_structure);

  thisyear = time_structure.tm_year;
  if (thisyear < 70) thisyear += 100;
  if (thisyear < 1900) thisyear += 1900;
  toabs (&utc, thisyear,
         time_structure.tm_mon,     time_structure.tm_mday,
         time_structure.tm_hour,    time_structure.tm_min,
         time_structure.tm_sec);

  gmt = utc.day;                   /* UTC time of year */

  /* Compute julian date and time of year */

  fjd = juldat(&utc) - gmt - JDEPOCH(STANDARD_EPOCH);
  *toy = fjd + gmt;
  *gha = 360.0 * modf(gmt, &dum); 
#if (DEBUG > 1)
  fprintf(stderr,"toy=%f, gha=%f\n",*toy,*gha);
#endif

  *ob = 23.439 - (*toy)*4.0e-7;

  sml = fmod (280.460 + 0.9856474*(*toy), 360.0);
  if (sml < 0.0) sml += 360.0;

  gst = fmod (sml + 180.0 + (*gha), 360.0);
  *lst = fmod (360.0 + gst + lon, 360.0);
}

/* ---------------------------------------------------------- */

static void get_solar_v(double toy,double ob,double* v)
{
  double sml,g,stl;

  /* Compute solar mean longitude (sidereal time) */
  sml = fmod (280.460 + 0.9856474*toy, 360.0);
  if (sml < 0.0) sml += 360.0;

  /* Compute solar true longitude */
  g = 357.528 + 0.9856003*toy;
  stl = sml + 1.915*sin(g/RADS) + 0.020*sin(2.0*g/RADS);

  /* Compute solar position vector */
  vmake (v, cos(stl/RADS), sin(stl/RADS), 0.0);
  vrot (v, ob/RADS, 1);
}

/* ---------------------------------------------------------- */
/* global functions */
/* ---------------------------------------------------------- */

void get_sun(time_t t,double lon,double lat,
             double* ra,double* dec,double* alt,double* az)
{
  double toy,gha,ob,lst,v[3],a,z;

  get_times(t,lon,&toy,&gha,&ob,&lst);
  get_solar_v(toy,ob,v);

  sphere (&a, &z, v);               /* Extract solar position */
  *ra  = a*RADS/15.0;
  *dec = z*RADS;
#if (DEBUG > 1)
  fprintf(stderr,"R.A.(sun)=%f, DEC(sun)=%f\n",*ra,*dec);
#endif

  /* Rotate the solar position into horizon coordinates */
  vrot (v, (-lst)/RADS, 3);
#if 0 /* not used later */
  { double tst,mst;
    sphere (&a, &z, v);
    tst = fmod (540.0 - RADS*a, 360.0);      /* True solar time */
    mst = fmod (720.0+gha+lon, 360.0);
  }
#endif
  vrot (v, (lat-90.0)/RADS, 2);       /* Geodetic */

  /* Extract the current altitude and azimuth */
  sphere (&a, &z, v);
  *az = fmod (540.0 - RADS * a, 360.0);
  *alt = refract (RADS * z);
#if (DEBUG > 1)
  fprintf(stderr,"AZ(sun)=%f, ALT(sun)=%f\n",*az,*alt);
#endif
}

/* ---------------------------------------------------------- */

static void get_lunar_v(double toy,double ob,double lst,double lat,double* mv)
{
  double tt,mlon,mlat,mpi,mr,ev[3],ff,c;

  tt = toy / 36525.0;

  mlon = 218.32 + 481267.883*tt
                + 6.29*sin((134.9+477198.85*tt)/RADS)
                - 1.27*sin((259.2-413335.38*tt)/RADS)
                + 0.66*sin((235.7+890534.23*tt)/RADS)
                + 0.21*sin((269.9+954397.70*tt)/RADS)
                - 0.19*sin((357.5+ 35999.05*tt)/RADS)
                - 0.11*sin((186.6+966404.05*tt)/RADS);
  mlat =
                  5.13*sin(( 93.3+483202.03*tt)/RADS)
                + 0.28*sin((228.2+960400.87*tt)/RADS)
                - 0.28*sin((318.3+  6003.18*tt)/RADS)
                - 0.17*sin((217.6-407332.20*tt)/RADS);
  mpi = 0.9508
                + 0.0518*cos((134.9+477198.85*tt)/RADS)
                + 0.0095*cos((259.2-413335.38*tt)/RADS)
                + 0.0078*cos((235.7+890534.23*tt)/RADS)
                + 0.0028*cos((269.9+954397.70*tt)/RADS);
  mr = 1.0 / sin(mpi/RADS);

  /* Make moon position vector in equitorial coordinates */
  vmake(mv, mr*cos(mlat/RADS), 0.0, mr*sin(mlat/RADS));
  vrot(mv, mlon/RADS, 3);
  vrot(mv, ob/RADS, 1);

  /* Correct for geocentric parallax */
  /* Need geocentric latitude */
  ff = square(1.0 - 1.0/298.25);
  c = 1.0/sqrt(square(cos(lat/RADS))+ff*square(sin(lat/RADS)));
  vmake (ev, c*cos(lat/RADS), 0.0, ff*c*sin(lat/RADS));
  vrot (ev, lst/RADS, 3);
  vdif (mv, ev);
}

/* ---------------------------------------------------------- */

void get_moon(time_t t,double lon,double lat,
              double* ra,double* dec,double* alt,double* az,double* phase)
{
  double toy,gha,ob,lst,mv[3],a,z;

  get_times(t,lon,&toy,&gha,&ob,&lst);
  get_lunar_v(toy,ob,lst,lat,mv);

  sphere (ra, dec, mv);                /* Extract lunar position */
  *ra  *= RADS/15.0;
  *dec *= RADS;
#if (DEBUG > 1)
  fprintf(stderr,"R.A.(moon)=%f, DEC(moon)=%f\n",*ra,*dec);
#endif

#if 0 /* mlon, mlat are not used later */
  { double x[3];  /* Find parallax affected longitude and latitude */
    vcopy(x, mv);                      /* mv used below */
    vrot(x, -ob/RADS, 1);
    sphere (&mlon, &mlat, x);
    mlon *= RADS;
    mlat *= RADS;
  }
#endif

  /* Rotate topocentric vector into horizon system */
  vrot(mv, (-lst)/RADS, 3);
  vrot(mv, (lat-90.0)/RADS, 2);        /* Geodetic */

  /* Extract the current altitude and azimuth */
  sphere (&a, &z, mv);
  *az = fmod (540.0 - RADS * a, 360.0);
  *alt = refract (RADS * z);
#if (DEBUG > 1)
  fprintf(stderr,"AZ(moon)=%f, ALT(moon)=%f\n",*az,*alt);
#endif

  if (phase) { double v[3],r;           /* calculate moon phase */
    get_solar_v(toy,ob,v);
    vrot (v, (-lst)/RADS, 3);
    vrot (v, (lat-90.0)/RADS, 2);       /* Geodetic */
    r = vnorm(mv);                      /* Compensate for dif above */
    vxs(mv,1.0/r);
    *phase = RADS * acos(vdot(v,mv));
#if (DEBUG > 1)
    fprintf(stderr,"PHASE(moon)=%f",*phase);
#endif
  }
}

/* ---------------------------------------------------------- */
/* ---------------------------------------------------------- */
/* ---------------------------------------------------------- */

