/* ---------------------------------------------------------------- 
 *
 * zwo.c
 *
 * ---------------------------------------------------------------- */

#ifndef DEBUG
#define DEBUG      0                   /* debug level */
#endif

/* ---------------------------------------------------------------- */

#define _REENTRANT

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <assert.h>

#include "zwo.h"
#include "utils.h"
#include "tcpip.h"

/* ---------------------------------------------------------------- */

#define PREFUN     __func__

/* ---------------------------------------------------------------- */

ZwoStruct* zwo_create(const char* host,int port)
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,host,port);
#endif

  ZwoStruct *self = (ZwoStruct*)malloc(sizeof(ZwoStruct));
  if (!self) return NULL;

  strcpy(self->host,host);
  self->port = port;
  self->sock = -1;                     /* connection closed */
  strcpy(self->serverVersion,"0");
  strcpy(self->model,"ZWO");
  self->cookie = 0;
  self->maxWidth = self->maxHeight = 0;
  self->hasCooler = self->isColor = 0;
  self->tempSensor = self->percent = 0.0f;
  self->aoiW = self->aoiH = 0;
  self->startX = self->startY = 0;
  self->expTime = 0.025f;
  self->gain = 0;
  self->offset = 0;
  self->bits = 0;
  self->bitDepth = 0;
  self->filter = self->nfilters = 0;

  pthread_mutex_init(&self->ioLock,NULL);

  return self;
}

/* ---------------------------------------------------------------- */

static int zwo_request(ZwoStruct* self,const char* cmd,char* res,int tout)
{
  char command[128],answer[128];

  assert(strlen(cmd) < sizeof(command)+2);
  assert(self->sock >= 0);
  assert(pthread_mutex_trylock(&self->ioLock));  /* must fail here */

  sprintf(command,"%s\n",cmd);
  int err = TCPIP_Request3(self->sock,command,answer,sizeof(answer),tout);
  if (!err) { if (res) strcpy(res,answer); }

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_connect(ZwoStruct* self)
{
  int  err=0,sock=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s:%d)\n",PREFUN,self->host,self->port);
#endif
  assert(self);
  assert(self->sock < 0);

  pthread_mutex_lock(&self->ioLock);
  if (!err) {
    sock = TCPIP_CreateClientSocket(self->host,self->port,&err);
#if (DEBUG > 1)
    printf("%s(%s:%d): sock=%d\n",PREFUN,self->host,self->port,sock);
#endif
    if (!err) self->sock = sock;
  }
  if (!err) { char buf[128];
    err = zwo_request(self,"version",buf,5);
    if (!err) { time_t st;
      assert(strlen(buf) < sizeof(self->serverVersion));
      sscanf(buf,"%s %u %ld",self->serverVersion,&self->cookie,&st);
    }
    if (err) { close(sock); self->sock = -1; }
  }
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_open(ZwoStruct* self) /* bit-depth added NEW v0016 */
{
  int  err=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s:%d)\n",PREFUN,self->host,self->port);
#endif
  assert(self);

  if (self->sock < 0) err = zwo_connect(self);

  if (!err) { char buf[128];
    pthread_mutex_lock(&self->ioLock);
    err = zwo_request(self,"open",buf,10);
    if (!err) {
      sscanf(buf,"%d %d %d %d %d %s",&self->maxWidth,&self->maxHeight,
             &self->hasCooler,&self->isColor,&self->bitDepth,self->model);
    }
    pthread_mutex_unlock(&self->ioLock);
  }

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_setup(ZwoStruct* self,int x,int y,int w,int h,int bin,int bits)
{
  int  err=0,b,p;
  char cmd[128],buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%d,%d,%d,%d,%d)\n",PREFUN,x,y,w,h,bin,bits);
#endif
  assert(self);
  assert(self->sock >= 0);

  pthread_mutex_lock(&self->ioLock);

#if 0 // unused
  sprintf(cmd,"offtime %ld",cor_time(0));
  if (!err) err = zwo_request(self,cmd,buf,5);
#endif

  self->bits = bits;
  sprintf(cmd,"setup %d %d %d %d %d %d",x,y,w,h,bin,bits);
  if (!err) err = zwo_request(self,cmd,buf,5);
  if (!err) sscanf(buf,"%d %d %d %d %d %d",&x,&y,
                   &self->aoiW,&self->aoiH,&b,&p);

  sprintf(cmd,"exptime");
  if (!err) err = zwo_request(self,cmd,buf,5);
  if (!err) self->expTime = atof(buf);

  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_disconnect(ZwoStruct* self)
{
  int err=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif
  assert(self);

  pthread_mutex_lock(&self->ioLock);
  if (self->sock > 0) err = close(self->sock);
  self->sock = -1;
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_close(ZwoStruct* self)
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif
  assert(self);

  if (self->sock < 0) return 0;

  pthread_mutex_lock(&self->ioLock);
  int err = zwo_request(self,"close",NULL,5);
  pthread_mutex_unlock(&self->ioLock);
  zwo_disconnect(self);

  return err;
}

/* ---------------------------------------------------------------- */

void zwo_free(ZwoStruct* self)
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif
  assert(self);

  if (self->sock > 0) zwo_close(self);
  pthread_mutex_destroy(&self->ioLock);
  free((void*)self);
}

/* ---------------------------------------------------------------- */

int zwo_temperature(ZwoStruct *self)
{
  char buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif
  assert(self->sock >= 0);

  pthread_mutex_lock(&self->ioLock);
  int err = zwo_request(self,"tempcon",buf,5);
  pthread_mutex_unlock(&self->ioLock);
  if (!err) { double t,p;
    if (sscanf(buf,"%lf %lf",&t,&p) == 2) {
      self->tempSensor = (float)t;
      self->percent    = (float)p;
    }
  }

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_tempcon(ZwoStruct* self,int value)
{
  char cmd[128],buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,value);
#endif

  sprintf(cmd,"tempcon %d",value);
  pthread_mutex_lock(&self->ioLock);
  int err = zwo_request(self,cmd,buf,5);
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_exptime(ZwoStruct* self,double value)
{
  char cmd[128],buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%f)\n",PREFUN,value);
#endif

  value = fmin(30.0,fmax(0.0001,value));
  sprintf(cmd,"exptime %.6f",value);
  pthread_mutex_lock(&self->ioLock);
  int err = zwo_request(self,cmd,buf,5);
  pthread_mutex_unlock(&self->ioLock);
  if (!err) self->expTime = value;

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_gain(ZwoStruct* self,int value)
{
  char cmd[128],buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,value);
#endif

  sprintf(cmd,"gain %d",value);
  pthread_mutex_lock(&self->ioLock);
  int err = zwo_request(self,cmd,buf,5);
  pthread_mutex_unlock(&self->ioLock);
  if (!err) self->gain = value;

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_offset(ZwoStruct* self,int value)
{
  char cmd[128],buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,value);
#endif

  sprintf(cmd,"offset %d",value);
  pthread_mutex_lock(&self->ioLock);
  int err = zwo_request(self,cmd,buf,5);
  pthread_mutex_unlock(&self->ioLock);
  if (!err) self->offset = value;

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_filter(ZwoStruct* self,int value,int wait)
{
  int  err=0;
  char cmd[128],buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,value);
#endif

  pthread_mutex_lock(&self->ioLock);
  if (self->nfilters == 0) {
    err = zwo_request(self,"filters",buf,5);
    if (!err) { 
      self->nfilters = atoi(buf);
      if (self->nfilters <= 0) err = -3;
    }
  }
  if (!err) {
    value = fmin(self->nfilters-1,fmax(0,value));
    sprintf(cmd,"filter %d",value);
    err = zwo_request(self,cmd,buf,5);
    if (!err) {
      if (strncmp(buf,"-E",2)) {
        self->filter = atoi(buf);
      } else {
        char *p = strchr(buf,'='); if (p) err = atoi(p+1);
      }
    }
  }
  if (!err && wait) { int timeout=12,f=-1; // ?Povilas
    strcpy(cmd,"filter");
    do {
      sleep(1); timeout -= 1;
      err = zwo_request(self,cmd,buf,5); if (err) break;
      f = atoi(buf); if (f == value) break;
    } while (timeout > 0);
    if (f != value) err = -4;
  }
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_flip(ZwoStruct* self,int x,int y)
{
  char cmd[128],buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%d,%d)\n",PREFUN,self,x,y);
#endif

  sprintf(cmd,"flip %d %d",x,y);
  pthread_mutex_lock(&self->ioLock);
  int err = zwo_request(self,cmd,buf,5);
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_expose(ZwoStruct* self)
{
  char cmd[128],buf[128];

  sprintf(cmd,"expose");
  pthread_mutex_lock(&self->ioLock);
  int err = zwo_request(self,cmd,buf,5);
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int zwo_data(ZwoStruct* self,void* data,int timeout)
{
  int  err=0;
  ssize_t r;
  char cmd[128],buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p,%d)\n",PREFUN,data,timeout);
#endif

  sprintf(cmd,"data");
  pthread_mutex_lock(&self->ioLock);
  double t1 = walltime(0);
  do {
    msleep(350);
    err = zwo_request(self,cmd,buf,5);
    if (err) break;
    if (!strncmp(buf,"-E",2)) { 
      err = -1;
    } else { int i,n;
      int size = atoi(buf);
      assert(size == self->aoiW*self->aoiH*self->bits/8);
      for (i=0,n=0; i<size/42; i++) {
        r = TCPIP_Recv(self->sock,(char*)data+n,size-n,2);
        if (r <= 0) break;
        n += r; if (n >= size) break;
      }
      if (n != size) err = -2;
      break;
    }
  } while (walltime(0)-t1 < timeout);
  pthread_mutex_unlock(&self->ioLock);

  if (!err) { int i;
    int npix = self->aoiW*self->aoiH;
    if (self->bits == 16) {            /* shift bits */
      if (self->bitDepth == 12) {      /* NEW v0016 */
        u_short *p = (u_short*)data;
        for (i=0; i<npix; i++,p++) *p = *p >> 4;
      }
    } else 
    if (self->bits == 24) {            /* swap RGB */
      u_char *p=(u_char*)data,*q=(u_char*)data+2,h;
      for (i=0; i<npix; i++,p+=3,q+=3) { h = *p; *p = *q; *q = h; }
    }
  }

  return err;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

